import functools
import gc
import urllib.parse
from pathlib import Path
from typing import Tuple, Any

import rocksdb
from rocksdb import errors

from kona.key_value_store import KeyValueStoreError
from kona.key_value_store import KeyValueStoreWriteBatch, KeyValueStoreCancelableWriteBatch, KeyValueStore
from kona.key_value_store import _validate_args_bytes, _validate_args_bytes_without_first

rocksdb_exceptions = [errors.NotFound,
                      errors.Corruption,
                      errors.NotSupported,
                      errors.InvalidArgument,
                      errors.RocksIOError,
                      errors.MergeInProgress,
                      errors.Incomplete]


def _error_convert(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            if type(e) in rocksdb_exceptions:
                raise KeyValueStoreError(e)
            raise e

    return _wrapper


class _KeyValueStoreWriteBatchRocksDB(KeyValueStoreWriteBatch):
    def __init__(self, db: rocksdb.DB, sync: bool):
        self._db = db
        self._sync = sync
        self._batch = self._new_batch()

    @_error_convert
    def _new_batch(self):
        return rocksdb.WriteBatch()

    @_validate_args_bytes_without_first
    @_error_convert
    def put(self, key: bytes, value: bytes):
        self._batch.put(key, value)

    @_validate_args_bytes_without_first
    @_error_convert
    def delete(self, key: bytes):
        self._batch.delete(key)

    @_error_convert
    def clear(self):
        self._batch.clear()

    @_error_convert
    def write(self):
        self._db.write(self._batch, sync=self._sync)


class _KeyValueStoreCancelableWriteBatchRocksDB(KeyValueStoreCancelableWriteBatch):
    def __init__(self, store: KeyValueStore, db: rocksdb.DB, sync: bool):
        super().__init__(store, sync=sync)
        self._touched_keys = set()
        self._db = db
        self._snapshot = db.snapshot()

    def _touch(self, key: bytes):
        self._touched_keys.add(key)

    def _get_original_touched_item(self):
        for key in self._touched_keys:
            try:
                yield key, self._db.get(key, snapshot=self._snapshot)
            except KeyError:
                return key, None

    def clear(self):
        super().clear()
        self._touched_keys.clear()

    def close(self):
        self._snapshot.close()
        self._snapshot = None


class KeyValueStoreRocksDB(KeyValueStore):
    def __init__(self, uri: str, **kwargs):
        uri_obj = urllib.parse.urlparse(uri)
        if uri_obj.scheme != 'file':
            raise ValueError(f"Support file path URI only (ex. file:///xxx/xxx). uri={uri}")
        self._path = f"{(uri_obj.netloc if uri_obj.netloc else '')}{uri_obj.path}"
        self._db = self._new_db(self._path, **kwargs)

    @_error_convert
    def _new_db(self, path, **kwargs) -> rocksdb.DB:
        return rocksdb.DB(path, rocksdb.Options(**kwargs))

    @_validate_args_bytes_without_first
    @_error_convert
    def get(self, key: bytes, *, default=None, **kwargs) -> bytes:
        if default is not None:
            _validate_args_bytes(default)

        result = self._db.get(key, **kwargs) or default
        if result is None:
            raise KeyError(f"Has no value of key({key})")
        return result

    @_validate_args_bytes_without_first
    @_error_convert
    def put(self, key: bytes, value: bytes, *, sync=False, **kwargs):
        self._db.put(key, value, sync=sync, **kwargs)

    @_validate_args_bytes_without_first
    @_error_convert
    def delete(self, key: bytes, *, sync=False, **kwargs):
        self._db.delete(key, sync=sync, **kwargs)

    @_error_convert
    def close(self):
        if self._db:
            del self._db
            gc.collect()
            self._db = None

    @_error_convert
    def destroy_store(self):
        self.close()

        def rm_tree(path: Path):
            for child in path.iterdir():
                if child.is_file():
                    child.unlink()
                else:
                    rm_tree(child)
            path.rmdir()

        rm_tree(Path(self._path))

    @_validate_args_bytes_without_first
    @_error_convert
    def key_may_exist(self, key: bytes) -> Tuple[bool, Any]:
        return self._db.key_may_exist(key, fetch=True)

    @_error_convert
    def WriteBatch(self, sync=False) -> KeyValueStoreWriteBatch:
        return _KeyValueStoreWriteBatchRocksDB(self._db, sync=sync)

    @_error_convert
    def CancelableWriteBatch(self, sync=False) -> KeyValueStoreCancelableWriteBatch:
        return _KeyValueStoreCancelableWriteBatchRocksDB(self, self._db, sync=sync)

    @_error_convert
    def Iterator(self, start_key: bytes = None, stop_key: bytes = None, include_value: bool = True, **kwargs):
        """Get Iterator

        :param start_key:
        :param stop_key:
        :param include_value:  # This parameter is not handled in rocksdb.
        :param kwargs:  # This parameter is not handled in rocksdb.
        :return:
        """
        if 'start' in kwargs or 'stop' in kwargs:
            raise ValueError("Use start_key and stop_key arguments instead of start and stop arguments")

        it = self._db.iteritems()
        try:
            it.seek(start_key)
        except TypeError:
            it.seek_to_first()

        for key, value in it:
            yield key, value
            if stop_key and stop_key == key:
                return



"""Test KeyValueStore class"""
import pytest

from kona.key_value_store import KeyValueStoreError, KeyValueStore


class TestKeyValueStore:
    store_types = ['dict', 'rocksdb']

    def _get_test_items(self, count: int = 5):
        test_items = dict()
        for i in range(1, count + 1):
            key = bytes(f"test_key_{i}", encoding='utf-8')
            value = bytes(f"test_value_{i}", encoding='utf-8')
            test_items[key] = value
        return test_items

    def _new_store(self, uri, store_type=None, create_if_missing=True):
        try:
            if store_type == KeyValueStore.STORE_TYPE_DICT:
                from kona.key_value_store_dict import KeyValueStoreDict
                print(f"New KeyValueStore. store_type={store_type}, uri={uri}")
                return KeyValueStoreDict()

            return KeyValueStore.new(uri, store_type=store_type, create_if_missing=create_if_missing)
        except KeyValueStoreError as e:
            print(f"Doesn't need to clean the store. uri={uri}, e={e}")

        return KeyValueStore.new(uri, store_type=store_type, create_if_missing=True)

    @pytest.mark.parametrize("store_type", store_types, ids=store_types)
    def test_key_value_store_basic(self, store_type):
        test_items = self._get_test_items(5)
        print(f"test_items={test_items}")

        store = self._new_store("file://./key_value_store_test_basic", store_type=store_type)

        #
        # put/get
        #

        for key, value in test_items.items():
            store.put(key, value)
            assert store.get(key) == value

        with pytest.raises(KeyError):
            store.get(b'unknown_key')

        assert store.get(b'unknown_key', default=b'test_default_value') == b'test_default_value'

        kwargs = {}

        if store_type == 'dict':
            container = tuple(test_items.keys())
        else:
            kwargs.update({
                'start_key': b'test_key_2',
                'stop_key': b'test_key_4'
            })
            container = (b'test_key_2', b'test_key_3', b'test_key_4')
        expect_count = len(container)

        count = 0
        for key, value in store.Iterator(**kwargs):
            assert key in container
            count += 1
        assert count == expect_count

        count = 0
        for key, value in store.Iterator(**kwargs):
            assert key in container
            count += 1
        assert count == expect_count

        count = 0
        for key, value in store.Iterator(**kwargs):
            assert key in container
            count += 1
        assert count == expect_count

        #
        # delete
        #

        del_key = b'test_key_2'
        del test_items[del_key]
        store.delete(del_key)
        with pytest.raises(KeyError):
            store.get(del_key)

        count = 0
        for key, value in store.Iterator():
            print(f"DB iterator: key={key}, value={value}")
            assert value == test_items[bytes(key)]
            count += 1
        print(f"Count after {del_key} has been deleted={count}")
        assert count == len(test_items)

        store.destroy_store()

    @pytest.mark.parametrize("store_type", store_types, ids=store_types)
    def test_key_value_store_write_batch(self, store_type):
        store = self._new_store("file://./key_value_store_test_write_batch", store_type=store_type)

        batch = store.WriteBatch()
        batch.put(b'test_key_1', b'test_value_1')
        batch.put(b'test_key_2', b'test_value_2')

        with pytest.raises(KeyError):
            store.get(b'test_key_1')
        with pytest.raises(KeyError):
            store.get(b'test_key_2')

        batch.write()
        assert store.get(b'test_key_1') == b'test_value_1'
        assert store.get(b'test_key_2') == b'test_value_2'

        store.destroy_store()

    @pytest.mark.parametrize("store_type", store_types, ids=store_types)
    def test_key_value_store_cancelable_write_batch(self, store_type):
        test_items = self._get_test_items(5)

        store = self._new_store("file://./key_value_store_test_cancelable_write_batch", store_type=store_type)

        for key, value in test_items.items():
            store.put(key, value)

        cancelable_batch = store.CancelableWriteBatch()
        cancelable_batch.put(b'cancelable_key_1', b'cancelable_value_1')
        cancelable_batch.put(b'test_key_2', b'edited_test_value_2')
        cancelable_batch.put(b'cancelable_key_2', b'cancelable_value_2')
        cancelable_batch.put(b'test_key_4', b'edited_test_value_4')
        cancelable_batch.write()

        edited_test_items = test_items.copy()
        edited_test_items[b'cancelable_key_1'] = b'cancelable_value_1'
        edited_test_items[b'test_key_2'] = b'edited_test_value_2'
        edited_test_items[b'cancelable_key_2'] = b'cancelable_value_2'
        edited_test_items[b'test_key_4'] = b'edited_test_value_4'

        count = 0
        print("before cancel")
        for key, value in store.Iterator():
            print(f"Edited DB iterator: key={key}, value={value}")
            assert value == edited_test_items[bytes(key)]
            count += 1
        print(f"Count after cancelable_batch has been written={count}")

        assert count == len(edited_test_items)

        cancelable_batch.cancel()
        count = 0
        print("after cancel")
        for key, value in store.Iterator():
            print(f"Original DB iterator: key={key}, value={value}")
            assert value == test_items[bytes(key)]
            count += 1
        print(f"Original count={count}")
        assert count == len(test_items)

        store.destroy_store()



"""KeyValueStoreDict classes are components for development"""

import functools
from typing import Tuple, Any

from kona.key_value_store import KeyValueStoreError
from kona.key_value_store import KeyValueStoreWriteBatch, KeyValueStoreCancelableWriteBatch, KeyValueStore
from kona.key_value_store import _validate_args_bytes, _validate_args_bytes_without_first


def _error_convert(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RuntimeError as e:
            raise KeyValueStoreError(e)

    return _wrapper


class _KeyValueStoreWriteBatchDict(KeyValueStoreWriteBatch):
    def __init__(self, store_items: dict):
        self._store_items = store_items
        self._batch_items = dict()

    @_validate_args_bytes_without_first
    @_error_convert
    def put(self, key: bytes, value: bytes):
        self._batch_items[key] = value

    @_validate_args_bytes_without_first
    @_error_convert
    def delete(self, key: bytes):
        self._batch_items[key] = None

    @_error_convert
    def clear(self):
        self._batch_items.clear()

    @_error_convert
    def write(self):
        for key, value in self._batch_items.items():
            if value is None:
                try:
                    del self._store_items[key]
                except KeyError:
                    pass
            else:
                self._store_items[key] = value


class _KeyValueStoreCancelableWriteBatchDict(KeyValueStoreCancelableWriteBatch):
    def __init__(self, store: KeyValueStore, store_items: dict):
        super().__init__(store)
        self._store_items = store_items
        self._original_items = dict()

    def _touch(self, key: bytes):
        if key in self._original_items:
            return

        try:
            value = self._store_items[key]
        except KeyError:
            value = None
        self._original_items[key] = value

    def _get_original_touched_item(self):
        for key, value in self._original_items.items():
            yield key, value

    def clear(self):
        super().clear()
        self._original_items.clear()

    def close(self):
        self._original_items: dict = None


class KeyValueStoreDict(KeyValueStore):
    """KeyValueStoreDict class is just for development"""

    def __init__(self, **kwargs):
        self._store_items = dict()

    @_validate_args_bytes_without_first
    @_error_convert
    def get(self, key: bytes, *, default=None, **kwargs) -> bytes:
        if default is not None:
            _validate_args_bytes(default)

        result = self._store_items.get(key, default)
        if result is None:
            raise KeyError(f"Has no value of key({key})")
        return result

    @_validate_args_bytes_without_first
    @_error_convert
    def put(self, key: bytes, value: bytes, *, sync=False, **kwargs):
        self._store_items[key] = value

    @_validate_args_bytes_without_first
    @_error_convert
    def delete(self, key: bytes, *, sync=False, **kwargs):
        try:
            del self._store_items[key]
        except KeyError:
            pass

    @_error_convert
    def close(self):
        if not self._store_items:
            self._store_items = None

    @_error_convert
    def destroy_store(self):
        self.close()

    @_validate_args_bytes_without_first
    @_error_convert
    def key_may_exist(self, key: bytes) -> Tuple[bool, Any]:
        pass

    @_error_convert
    def WriteBatch(self, sync=False) -> KeyValueStoreWriteBatch:
        return _KeyValueStoreWriteBatchDict(self._store_items)

    @_error_convert
    def CancelableWriteBatch(self, sync=False) -> KeyValueStoreCancelableWriteBatch:
        return _KeyValueStoreCancelableWriteBatchDict(self, self._store_items)

    @_error_convert
    def Iterator(self, start_key: bytes = None, stop_key: bytes = None, include_value: bool = True, **kwargs):
        if start_key is not None or stop_key is not None:
            raise ValueError("Unsupported arguments which are start_key and stop_key")

        if include_value:
            return self._store_items.items()
        else:
            return self._store_items.keys()



import os
from multiprocessing import Process
from pathlib import Path
from shutil import rmtree
from time import sleep

import pytest
import rocksdb
from rocksdb.errors import RocksIOError


class TestDBPath:
    leveldb = 'test_leveldb'
    rocksdb = 'test_rocksdb'


def print_process_info(title):
    print("\n--------------------------")
    print(title)
    print('module name:', __name__)
    print('parent process:', os.getppid())
    print('process id:', os.getpid())


def f_rocksdb(name: str):
    print_process_info(f'function f_rocksdb({name})')
    data = b'test_test_test'

    try:
        db = rocksdb.DB(TestDBPath.rocksdb, rocksdb.Options(create_if_missing=True))
        db.put(b'name', data)
    except RocksIOError:
        sleep(0.1)
        db = rocksdb.DB(TestDBPath.rocksdb, rocksdb.Options(create_if_missing=True), read_only=True)
        if data != db.get(b"name"):
            exit(-1)

    sleep(1)


def delete_test_db_dirs():
    db_paths = [Path(f"./{TestDBPath.leveldb}"), Path(f"./{TestDBPath.rocksdb}")]

    for db_path in db_paths:
        if db_path.exists():
            print(f'delete DB({db_path.resolve()})')
            rmtree(db_path)


@pytest.fixture(autouse=True)
def run_around_tests():
    delete_test_db_dirs()
    yield
    delete_test_db_dirs()


class TestMultiProcessLevelDB:
    def test_multiprocessing_rocksdb(self):
        p1 = Process(target=f_rocksdb, args=('rocksdb_prop1',))
        p1.start()

        p2 = Process(target=f_rocksdb, args=('rocksdb_prop2',))
        p2.start()

        p1.join()
        p2.join()

        # RocksDB supports multiprocessing.
        assert p1.exitcode == 0
        assert p2.exitcode == 0



import abc
import functools
from typing import Union, Tuple, Any

from yirgachefe import config


class KeyValueStoreError(Exception):
    pass


class KeyValueStoreWriteBatch(abc.ABC):
    """For batch put and delete operations."""

    @abc.abstractmethod
    def put(self, key: bytes, value: bytes):
        """Add or modify a value of the key temporarily.

        :param key:
        :param value:
        """
        raise NotImplementedError("put() function is interface method")

    @abc.abstractmethod
    def delete(self, key: bytes):
        """Delete a record of the key temporarily.

        :param key:
        """
        raise NotImplementedError("delete() function is interface method")

    @abc.abstractmethod
    def clear(self):
        """Clear batch operations."""
        raise NotImplementedError("clear() function is interface method")

    @abc.abstractmethod
    def write(self):
        """Write batch put and delete operations."""
        raise NotImplementedError("write() function is interface method")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            self.write()


class KeyValueStoreCancelableWriteBatch(abc.ABC):
    """Cancelable batch put and delete operations.

    It can cancel operations after the operations has been written (called write() function).
    However, It just recover key/value pairs to before KeyValueStoreCancelableWriteBatch instance was made.

    # Store: key1/value1, key2/value2
    batch = key_value_store_instance.CancelableWriteBatch()
    batch.put(key1, value1-1)
    batch.delete(key2)
    batch.write()
    # Store: key1/value-1 (deleted key2)

    # (X) key_value_store_instance.put(key1, value1-2) # will be recovered (key1/value1) after batch is canceled.
    batch.cancel()

    # Store: key1/value1, key2/value2
    """

    def __init__(self, store: 'KeyValueStore', sync=False):
        self._store = store
        self._batch = self._store.WriteBatch(sync=sync)
        self._sync = sync

    def put(self, key: bytes, value: bytes):
        """Add or modify a value of the key temporarily.

        :param key:
        :param value:
        """
        self._batch.put(key, value)
        self._touch(key)

    def delete(self, key: bytes):
        """Delete a record of the key temporarily.

        :param key:
        """
        self._batch.delete(key)
        self._touch(key)

    def clear(self):
        """Clear batch operations."""
        self._batch.clear()

    def write(self):
        """Write batch put and delete operations."""
        self._batch.write()

    def cancel(self):
        """Cancel written operations."""
        batch = self._store.WriteBatch(sync=self._sync)
        for key, value in self._get_original_touched_item():
            if value is None:
                batch.delete(key)
            else:
                batch.put(key, value)
        batch.write()

    @abc.abstractmethod
    def close(self):
        """Close explicitly.

        Will be closed automatically when this instance is deleted.
        """
        raise NotImplementedError("close() function is interface method")

    @abc.abstractmethod
    def _touch(self, key: bytes):
        raise NotImplementedError("_touch() function is interface method")

    @abc.abstractmethod
    def _get_original_touched_item(self):
        # Children have to override function as generator. return key, value
        raise NotImplementedError("_get_touched_item() function is interface method")


class KeyValueStore(abc.ABC):
    STORE_TYPE_ROCKSDB = 'rocksdb'
    STORE_TYPE_DICT = 'dict'

    @staticmethod
    def new(uri: str, store_type: str = None, **kwargs) -> 'KeyValueStore':
        if store_type is None:
            store_type = config.DEFAULT_KEY_VALUE_STORE_TYPE

        if store_type == KeyValueStore.STORE_TYPE_ROCKSDB:
            from kona.key_value_store_rocksdb import KeyValueStoreRocksDB
            return KeyValueStoreRocksDB(uri, **kwargs)
        elif store_type == KeyValueStore.STORE_TYPE_DICT:
            raise ValueError("KeyValueStoreDict is just for development.")
        else:
            raise ValueError(f"store_name is invalid. store_type={store_type}")

    @abc.abstractmethod
    def get(self, key: bytes, *, default=None, **kwargs) -> bytes:
        """Get a value of the key

        :param key:
        :param default: default (bytes)
        :param kwargs:
        :return: a value of the key
        """
        raise NotImplementedError("get() function is interface method")

    @abc.abstractmethod
    def put(self, key: bytes, value: bytes, *, sync=False, **kwargs):
        """Add or modify a value of the key.

        :param key:
        :param value:
        :param sync:
        """
        raise NotImplementedError("put() function is interface method")

    @abc.abstractmethod
    def delete(self, key: bytes, *, sync=False, **kwargs):
        """Delete a record of the key.

        :param key:
        :param sync:
        """
        raise NotImplementedError("delete() function is interface method")

    @abc.abstractmethod
    def close(self):
        """Close explicitly.

        Will be closed automatically when this instance is deleted.
        """
        raise NotImplementedError("close() function is interface method")

    @abc.abstractmethod
    def destroy_store(self):
        """Destroy store data.

        If the data are files, the files may be deleted.
        """
        raise NotImplementedError("destroy_store() function is interface method")

    @abc.abstractmethod
    def key_may_exist(self, key: bytes) -> Tuple[bool, Any]:
        """If the key definitely does not exist in the database, then this method returns False, else True."""
        raise NotImplementedError("destroy_store() function is interface method")

    @abc.abstractmethod
    def WriteBatch(self, sync=False) -> KeyValueStoreWriteBatch:
        """Make a KeyValueStoreWriteBatch instance for this instance"""
        raise NotImplementedError("WriteBatch constructor is not implemented in KeyValueStore class")

    @abc.abstractmethod
    def CancelableWriteBatch(self, sync=False) -> KeyValueStoreCancelableWriteBatch:
        """Make a KeyValueStoreCancelableWriteBatch instance for this instance"""
        raise NotImplementedError("CancelableWriteBatch constructor is not implemented in KeyValueStore class")

    @abc.abstractmethod
    def Iterator(self, start_key: bytes = None, stop_key: bytes = None, include_value: bool = True, **kwargs):
        """Return iterator

        :param start_key: a start key (inclusive)
        :param stop_key: a stop key (inclusive)
        :param include_value: include value (for key, value in store_instance.Iterator(include_value=True):)
        :return: iterator
        """
        # TODO: make KeyValueStoreIterator class. Currently, Iterator can be used in for-loop.
        raise NotImplementedError("Iterator constructor is not implemented in KeyValueStore class")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def _validate_args_bytes(arg: Union[bytes, bytearray]):
    if not isinstance(arg, (bytes, bytearray)):
        raise ValueError(f"Argument type({type(arg)}) is not bytes. argument={arg}")


def _validate_args_bytes_without_first(func):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        for arg in args[1:]:
            _validate_args_bytes(arg)
        return func(*args, **kwargs)

    return _wrapper
